//      flashprog.c
//      
//      Copyright 2012 swausd
//      
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.
//
//      -----
//
//      This are some subroutines for programming the flash memory in the
//      Vivonic Fitness Planner, a PDA sytle device based on a Freescale
//      68EZ328 Dragonball cpu.
//      Flash memory is a Intel TE28F160-C3TA90
//      16 MBit organized in 16 Bit Words
//      Operations are done using 16 bit words. Only simplest error checking
//      is imolemented.
//      Note: erased memory cells contain 0xff. Only zeros are programmed.


#include "flashprog.h"
#include "serio.h"

#define DEBUG 1

// IMPORTANT: flashBaseAdr has to have volative attribute,
// so gcc dosn't optimize and dismisses some write statements

static uint16_t volatile *flashBaseAdr = 0;

// Sets starting address of memory in cpu address space. All following
// operations are done on relative addresses based on this base.

void flashSetBaseAdr(uint16_t *adr)
{
   flashBaseAdr = adr;
}


// Wait until ready flag is set.
// Note: no timeout implemented, so system may hang!
   
uint8_t flashWaitUntilReady(void)
{
   uint8_t status;
   
   do
   {
      status = *flashBaseAdr;
   
#if DEBUG >= 2
	putStringToUart("flashWaitUntilReady: ");
	putHexByteToUart(status);
	putCrNlToUart();
#endif
   }
   while (!(status & FLASH_STATUS_WSMS));
   
   return status;
}


// Read and return current status byte of flash memory.
// Flash has to be in one of the special modes programming/erase ...
// Otherwise the contents of the flash memory at base address is read.

uint8_t flashGetStatus(void)
{
   return *flashBaseAdr;
}


// Read word at base address + offest as 16 bit word.
// offset has to be specified as byte offset and is 
// converted to an even word address by divied by 2.

uint16_t flashGetWord(uint32_t offset)
{
   offset = offset >> 1;
   
   return *(flashBaseAdr+offset);
}



// Write data as 16 bit word to base address + offest.
// offset has to be specified as byte offset and is 
// converted to an even word address by divied by 2.

void flashSetWord(uint16_t data, uint32_t offset)
{
   offset = offset >> 1;
   
   *(flashBaseAdr+offset) = data;
}


// Just a disguise for flashSetWord() used to write
// commands to the flash command register.

void flashSetCommand(uint16_t command, uint32_t offset)
{
   flashSetWord(command, offset);
}


// Erase a block of flash memory.
// See data sheet for valid block addresses
// offset has to be specified as byte offset and is 
// converted to an even word address by divied by 2.

uint8_t flashEraseBlock(uint32_t offset)
{
   uint8_t error = FLASH_NO_ERROR;
   uint8_t status;
   
#if DEBUG >= 1
	putStringToUart("flashEraseBlock\r\n");
#endif
   
   flashSetCommand(FLASH_CMD_ERASE_SETUP, offset);
   flashSetCommand(FLASH_CMD_ERASE_CONFIRM, offset);
   
   status = flashWaitUntilReady();
   
   if (status & (FLASH_STATUS_VPPS | FLASH_STATUS_ES))
      error = FLASH_ERASE_ERROR;
   else if (status & FLASH_STATUS_BLS)
      error = FLASH_BLOCK_LOCKED_ERROR;

   flashSetCommand(FLASH_CMD_CLEAR_STATUS, 0);
      
   return error;
}


// Write data as wordCount 16 bit words into flash starting at
// base address + offest.
// offset has to be specified as byte offset and is 
// converted to an even word address by divied by 2.

uint8_t flashWriteWord(uint32_t offset, uint16_t *data, uint32_t wordCount)
{
   uint8_t  error = FLASH_NO_ERROR;
   uint8_t  status;
   uint32_t index = 0;
   
#if DEBUG >= 1
	putStringToUart("flashWriteWord\r\n");
#endif
   
   if ((wordCount > 0) && (data != NULL))
   {
      do
      {
         flashSetCommand(FLASH_CMD_PROGRAM_SETUP, offset+(index << 1));
         flashSetWord(data[index], offset+(index << 1));

         status = flashWaitUntilReady();

         if (status & (FLASH_STATUS_VPPS | FLASH_STATUS_PS))
            error = FLASH_WRITE_ERROR;
         else if (status & FLASH_STATUS_BLS)
            error = FLASH_BLOCK_LOCKED_ERROR;
      }
      while (!error && (++index < wordCount));
   }
   else
      error = FLASH_PARAMETER_ERROR;
      
   flashSetCommand(FLASH_CMD_CLEAR_STATUS, 0);
      
   return error;
}


// reset flash memory to normal read operation

void flashNormalOperation(void)
{
#if DEBUG >= 1
	putStringToUart("flashNormalOperation\r\n");
#endif

   flashSetCommand(FLASH_CMD_READ_ARRAY, 0);
}


// Unlock flash memory block so it can be erased and written to.
// offset has to be specified as byte offset and is 
// converted to an even word address by divied by 2.

void flashUnlockBlock(uint32_t offset)
{
#if DEBUG >= 1
	putStringToUart("flashUnlockBlock\r\n");
#endif

	flashSetCommand(FLASH_CMD_CONFIG_SETUP, 0);
   flashSetCommand(FLASH_CMD_UNLOCK_BLOCK, offset);
}
